import tkinter as tk
import numpy as np
import time

from PIL import Image, ImageTk
from .node import *
from . import math


class Engine:
    def __init__(
        self,
        title: str = "TkEnginer",
        width: int = 1600,
        height: int = 900,
        fps: int = 60,
        fov: float = 90,
        near: float = 0.01,
        far: float = 100,
        clear_color: tuple[int, int, int] = (0, 0, 0),
        scene: Node = None
    ) -> None:

        self.window = tk.Tk()
        self.window.title(title)
        self.window.geometry(f"{width}x{height}")
        self.frame_time = 1000 / fps
        self.fov = fov
        self.near = near
        self.far = far
        self.clear_color = clear_color

        self.canvas = tk.Canvas(
            self.window,
            highlightthickness=0
        )
        self.canvas.pack(fill=tk.BOTH, expand=True)
        self.init(width, height)
        self.last_time = time.time()

        self.yaw = np.pi
        self.pitch = .0
        self.position = np.array([0, 0, 0], dtype=np.float32)

        self.scene = scene if scene is not None else Node()

        self.pressed_keys: set[str] = set()
        self.mouse: list[int] = None

        self.window.bind("<KeyPress>", self.key_pressed)
        self.window.bind("<KeyRelease>", self.key_released)
        self.window.bind("<ButtonPress>", self.button_pressed)
        self.window.bind("<ButtonRelease>", self.button_released)
        self.window.bind("<Motion>", self.mouse_moved)
        self.window.bind("<Configure>", self.window_resized)

    def init(self, width: int, height: int) -> None:
        self.width = width
        self.height = height
        self.projection_matrix = math.get_projection_matrix(
            self.fov,
            self.width, 
            self.height, 
            self.near, 
            self.far
        )
        self.buffer = np.zeros((self.height, self.width, 4), dtype=np.uint8)
        self.zbuffer = np.full((self.height, self.width), np.inf, dtype=np.float32)
        self.image = Image.fromarray(self.buffer, "RGBA")
        self.photo = ImageTk.PhotoImage(self.image)
        self.canvas.create_image(0, 0, image=self.photo, anchor="nw")

    def update(self, delta: float) -> None:
        pass

    def is_key_pressed(self, key: str) -> bool:
        return key in self.pressed_keys

    def get_mouse_position(self) -> list[int]:
        return self.mouse if self.mouse is not None else [0, 0]

    def run(self) -> None:
        self.loop()
        self.window.mainloop()

    def key_pressed(self, event: tk.Event) -> None:
        self.pressed_keys.add(event.keysym)

    def key_released(self, event: tk.Event) -> None:
        self.pressed_keys.discard(event.keysym)

    def button_pressed(self, event: tk.Event) -> None:
        self.pressed_keys.add(f"mouse_{event.num}")

    def button_released(self, event: tk.Event) -> None:
        self.pressed_keys.discard(f"mouse_{event.num}")

    def mouse_moved(self, event: tk.Event) -> None:
        self.mouse = [event.x, event.y]

    def window_resized(self, event: tk.Event) -> None:
        self.init(event.width, event.height)

    def loop(self) -> None: # TODO: shaders, lighting
        now = time.time()
        delta = now - self.last_time

        self.image.paste("black", (0, 0, self.width, self.height))
        self.buffer[:, :, :] = list(self.clear_color) + [255]
        self.zbuffer[:, :] = np.inf

        view_matrix = math.get_view_matrix(self.position, self.yaw, self.pitch)

        for node, global_transform in self.scene.traverse():
            node.update(delta)
            if node.mesh is None: continue
            vertices, indices, colors = node.mesh.get_data()
            mvp_matrix = self.projection_matrix @ view_matrix @ global_transform.get_matrix()

            vertices_clip = math.transform_vertices(vertices, mvp_matrix)
            screen_coords, w_coords = math.clip_to_screen(
                vertices_clip,
                self.width,
                self.height
            )

            for triangle in indices:
                w0, w1, w2 = w_coords[triangle[0], 0], w_coords[triangle[1], 0], w_coords[triangle[2], 0]
                if w0 <= 0 or w1 <= 0 or w2 <= 0:
                    continue

                p0, p1, p2 = screen_coords[triangle[0]], screen_coords[triangle[1]], screen_coords[triangle[2]]
                c0, c1, c2 = colors[triangle[0]], colors[triangle[1]], colors[triangle[2]]

                edge1 = p1 - p0
                edge2 = p2 - p0
                if float(edge1[0]) * float(edge2[1]) - float(edge1[1]) * float(edge2[0]) >= 0:
                    continue
                
                math.draw_triangle(self.buffer, self.zbuffer, p0, p1, p2, c0, c1, c2, w0, w1, w2)

        self.image = Image.fromarray(self.buffer, "RGBA")
        self.update(delta)
        self.photo.paste(self.image)

        self.last_time = now
        self.window.after(
            max(1, int(self.frame_time - 1000 * (time.time() - now))),
            self.loop
        )



import numpy as np


class Transform:
    def __init__(
        self,
        position: list[float] = None,
        rotation: list[float] = None,
        scale: list[float] = None
    ) -> None:
        position = position if position is not None else [0.0, 0.0, 0.0]
        rotation = rotation if rotation is not None else [0.0, 0.0, 0.0]
        scale = scale if scale is not None else [1.0, 1.0, 1.0]
        self.position = np.array(position, dtype=np.float32)
        self.rotation = np.array(rotation, dtype=np.float32)
        self.scale = np.array(scale, dtype=np.float32)

    def get_matrix(self) -> np.typing.NDArray[np.float32]:
        translation_matrix = np.identity(4, dtype=np.float32)
        translation_matrix[0, 3] = self.position[0]
        translation_matrix[1, 3] = self.position[1]
        translation_matrix[2, 3] = self.position[2]

        rotation_matrix = np.array([
            [np.cos(self.rotation[2]), -np.sin(self.rotation[2]), 0, 0],
            [np.sin(self.rotation[2]),  np.cos(self.rotation[2]), 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1]
        ], dtype=np.float32) @ np.array([
            [np.cos(self.rotation[1]), 0, np.sin(self.rotation[1]), 0],
            [0, 1, 0, 0],
            [-np.sin(self.rotation[1]), 0, np.cos(self.rotation[1]), 0],
            [0, 0, 0, 1]
        ], dtype=np.float32) @ np.array([
            [1, 0, 0, 0],
            [0, np.cos(self.rotation[0]), -np.sin(self.rotation[0]), 0],
            [0, np.sin(self.rotation[0]),  np.cos(self.rotation[0]), 0],
            [0, 0, 0, 1]
        ], dtype=np.float32)

        scaling_matrix = np.identity(4, dtype=np.float32)
        scaling_matrix[0, 0] = self.scale[0]
        scaling_matrix[1, 1] = self.scale[1]
        scaling_matrix[2, 2] = self.scale[2]

        return translation_matrix @ rotation_matrix @ scaling_matrix
    
    @classmethod
    def from_matrix(cls, matrix: np.typing.NDArray[np.float32]) -> "Transform":
        scale = np.linalg.norm(matrix[:3, :3], axis=0)

        rotation_matrix = matrix[:3, :3] / scale
        if np.abs(rotation_matrix[2, 0]) != 1:
            y = -np.arcsin(rotation_matrix[2, 0])
            cos_y = np.cos(y)
            x = np.arctan2(rotation_matrix[2, 1] / cos_y, rotation_matrix[2, 2] / cos_y)
            z = np.arctan2(rotation_matrix[1, 0] / cos_y, rotation_matrix[0, 0] / cos_y)
        else:
            z = 0
            if rotation_matrix[2, 0] == -1:
                y = np.pi / 2
                x = z + np.arctan2(rotation_matrix[0, 1], rotation_matrix[0, 2])
            else:
                y = -np.pi / 2
                x = -z + np.arctan2(-rotation_matrix[0, 1], -rotation_matrix[0, 2])
        rotation = np.array([x, y, z], dtype=np.float32)

        return cls(matrix[:3, 3], rotation, scale)
    
    def __matmul__(self, other: "Transform") -> np.ndarray:
        return Transform.from_matrix(self.get_matrix() @ other.get_matrix())



import numpy as np
import numba as nb


@nb.njit(cache=True)
def get_projection_matrix(
    fov: float,
    width: int,
    height: int,
    near: float,
    far: float
) -> np.ndarray:
    focal = 1 / np.tan(np.radians(fov) / 2)
    proj = np.zeros((4, 4), dtype=np.float32)
    proj[0, 0] = focal / (width / height)
    proj[1, 1] = focal
    proj[2, 2] = (far + near) / (near - far)
    proj[2, 3] = (2 * far * near) / (near - far)
    proj[3, 2] = -1
    return proj


@nb.njit(cache=True)
def get_camera_vectors(yaw: float, pitch: float) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    front = np.array([
        np.cos(pitch) * np.sin(yaw),
        np.sin(pitch),
        np.cos(pitch) * np.cos(yaw)
    ], dtype=np.float32)
    front /= np.linalg.norm(front)

    right = np.cross([0, 1, 0], front)
    right /= np.linalg.norm(right)

    up = np.cross(front, right)

    return front, right, up


@nb.njit(cache=True)
def get_view_matrix(position: np.ndarray, yaw: float, pitch: float) -> np.ndarray:
    front, right, up = get_camera_vectors(yaw, pitch)

    view = np.identity(4, dtype=np.float32)
    view[0, :3] = right
    view[1, :3] = up
    view[2, :3] = -front
    view[:3, 3] = -view[:3, :3] @ position
    return view


@nb.njit(cache=True)
def transform_vertices(vertices: np.ndarray, mvp_matrix: np.ndarray) -> np.ndarray:
    vertices_hom = np.concatenate(
        (vertices, np.ones((vertices.shape[0], 1), dtype=np.float32)),
        axis=1
    )
    vertices_clip = vertices_hom @ mvp_matrix.T
    return vertices_clip


@nb.njit(cache=True)
def clip_to_screen(vertices_clip: np.ndarray, width: int, height: int) -> tuple[np.ndarray, np.ndarray]:
    w_coords = vertices_clip[:, 3:4]
    vertices_ndc = vertices_clip[:, :3] / w_coords
    screen_coords = np.empty((len(vertices_ndc), 2), dtype=np.int32)
    screen_coords[:, 0] = ((vertices_ndc[:, 0] + 1) *
                           0.5 * width).astype(np.int32)
    screen_coords[:, 1] = (
        (1 - (vertices_ndc[:, 1] + 1) * 0.5) * height).astype(np.int32)
    return screen_coords, w_coords


@nb.njit(cache=True)
def lerp(a: float, b: float, t: float) -> float:
    return a + (b - a) * t

# TODO: color class


@nb.njit(cache=True)
def barycentric_weights(px, py, p0, p1, p2):
    v0x = p1[0] - p0[0]
    v0y = p1[1] - p0[1]
    v1x = p2[0] - p0[0]
    v1y = p2[1] - p0[1]

    d00 = v0x * v0x + v0y * v0y
    d01 = v0x * v1x + v0y * v1y
    d11 = v1x * v1x + v1y * v1y

    v2x = px - p0[0]
    v2y = py - p0[1]

    d20 = v2x * v0x + v2y * v0y
    d21 = v2x * v1x + v2y * v1y

    denom = d00 * d11 - d01 * d01
    if denom == 0:
        return -1.0, -1.0, -1.0

    v = (d11 * d20 - d01 * d21) / denom
    w = (d00 * d21 - d01 * d20) / denom
    u = 1.0 - v - w
    return u, v, w


@nb.njit(cache=True, parallel=True)
def draw_triangle(buffer, zbuffer, p0, p1, p2, c0, c1, c2, w0, w1, w2):
    w0 = 1 / w0
    w1 = 1 / w1
    w2 = 1 / w2
    height, width, channels = buffer.shape
    min_x = max(int(min(p0[0], p1[0], p2[0])), 0)
    max_x = min(int(max(p0[0], p1[0], p2[0])), width - 1)
    min_y = max(int(min(p0[1], p1[1], p2[1])), 0)
    max_y = min(int(max(p0[1], p1[1], p2[1])), height - 1)

    for y in nb.prange(min_y, max_y + 1):
        for x in range(min_x, max_x + 1):
            u, v, w = barycentric_weights(x + 0.5, y + 0.5, p0, p1, p2)
            if u >= 0 and v >= 0 and w >= 0:
                z = u / w0 + v / w1 + w / w2
                if z < zbuffer[y, x]:
                    zbuffer[y, x] = z
                    for ch in range(channels):
                        val = u * c0[ch] + v * c1[ch] + w * c2[ch]
                        if val < 0:
                            val = 0
                        elif val > 255:
                            val = 255
                        buffer[y, x, ch] = val



import numpy as np
import io


class Mesh:
    def __init__(
        self,
        vertices: list[list[float]],
        indices: list[list[int]],
        colors: list[list[int]] = None
    ) -> None:
        self.vertices = np.array(vertices, dtype=np.float32)
        self.indices = np.array(indices, dtype=np.uint32)
        if colors is not None:
            self.colors = np.array(colors, dtype=np.uint8)
        else:
            self.colors = np.full((len(vertices), 4), 255, dtype=np.uint8)

    def get_data(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        return self.vertices, self.indices, self.colors


class CubeMesh(Mesh):
    def __init__(self, colors: list[list[int]] = None) -> None:
        vertices = [
            [-0.5, -0.5, -0.5],
            [0.5, -0.5, -0.5],
            [0.5,  0.5, -0.5],
            [-0.5,  0.5, -0.5],
            [-0.5, -0.5,  0.5],
            [0.5, -0.5,  0.5],
            [0.5,  0.5,  0.5],
            [-0.5,  0.5,  0.5],
        ]
        indices = [
            [0, 1, 2], [0, 2, 3],
            [5, 4, 7], [5, 7, 6],
            [4, 0, 3], [4, 3, 7],
            [1, 5, 6], [1, 6, 2],
            [3, 2, 6], [3, 6, 7],
            [4, 5, 1], [4, 1, 0],
        ]
        super().__init__(vertices, indices, colors)


class PyramidMesh(Mesh):
    def __init__(self, colors: list[list[int]] = None) -> None:
        vertices = [
            [-0.5, -0.5, -0.5],
            [0.5, -0.5, -0.5],
            [0.0, -0.5, 0.5],
            [0.0, 0.5, 0.0],
        ]
        indices = [
            [0, 2, 1],
            [0, 1, 3],
            [1, 2, 3],
            [2, 0, 3],
        ]
        super().__init__(vertices, indices, colors)


class PyramidWithSquareBaseMesh(Mesh):
    def __init__(self, colors: list[list[int]] = None) -> None:
        vertices = [
            [-0.5, -0.5, -0.5],
            [0.5, -0.5, -0.5],
            [0.5, -0.5, 0.5],
            [-0.5, -0.5, 0.5],
            [0.0, 0.5, 0.0],
        ]
        indices = [
            [0, 3, 2],
            [0, 2, 1],
            [0, 1, 4],
            [1, 2, 4],
            [2, 3, 4],
            [3, 0, 4],
        ]
        super().__init__(vertices, indices, colors)


class SphereMesh(Mesh):
    def __init__(self, segments: int, colors: list[list[int]] = None) -> None:
        vertices = []
        indices = []

        for i in range(segments + 1):
            latitude = np.pi * i / segments
            for j in range(segments + 1):
                longitude = 2 * np.pi * j / segments
                vertices.append([
                    np.sin(latitude) * np.cos(longitude),
                    np.cos(latitude),
                    np.sin(latitude) * np.sin(longitude),
                ])

        for i in range(segments):
            for j in range(segments):
                x = i * (segments + 1) + j
                y = x + segments + 1
                indices.extend([
                    [x, y, x + 1],
                    [y, y + 1, x + 1]
                ])
        super().__init__(vertices, indices, colors)


class ConeMesh(Mesh):
    def __init__(self, segments: int, colors: list[list[int]] = None) -> None:
        top_vertex = [0.0, 0.5, 0.0]
        base_vertices = []

        for i in range(segments):
            angle = 2 * np.pi * i / segments
            x = 0.5 * np.cos(angle)
            z = 0.5 * np.sin(angle)
            base_vertices.append([x, -0.5, z])

        base_center = [0.0, -0.5, 0.0]
        vertices = [top_vertex] + base_vertices + [base_center]

        indices = []
        for i in range(segments):
            next_i = (i + 1) % segments
            indices.append([0, i + 1, next_i + 1])

        base_center_idx = len(vertices) - 1
        for i in range(segments):
            next_i = (i + 1) % segments
            indices.append([base_center_idx, next_i + 1, i + 1])

        super().__init__(vertices, indices, colors)


class CylinderMesh(Mesh):
    def __init__(self, segments: int, colors: list[list[int]] = None) -> None:
        top_center = [0.0, 0.5, 0.0]
        bottom_center = [0.0, -0.5, 0.0]
        top_vertices = []
        bottom_vertices = []

        for i in range(segments):
            angle = 2 * np.pi * i / segments
            x = 0.5 * np.cos(angle)
            z = 0.5 * np.sin(angle)
            top_vertices.append([x, 0.5, z])
            bottom_vertices.append([x, -0.5, z])

        vertices = [top_center] + top_vertices + \
            [bottom_center] + bottom_vertices

        indices = []
        for i in range(segments):
            next_i = (i + 1) % segments
            indices.append([0, i + 1, next_i + 1])

        bottom_center_idx = len(top_vertices) + 1
        bottom_vertex_start = bottom_center_idx + 1
        for i in range(segments):
            next_i = (i + 1) % segments
            indices.append(
                [bottom_center_idx, bottom_vertex_start + next_i, bottom_vertex_start + i])

        for i in range(segments):
            next_i = (i + 1) % segments
            top_idx = i + 1
            next_top_idx = next_i + 1
            bottom_idx = bottom_vertex_start + i
            next_bottom_idx = bottom_vertex_start + next_i

            indices.append([top_idx, bottom_idx, next_top_idx])
            indices.append([next_top_idx, bottom_idx, next_bottom_idx])

        super().__init__(vertices, indices, colors)


class PlaneMesh(Mesh):
    def __init__(self, colors: list[list[int]] = None) -> None:
        vertices = [
            [-0.5, 0.0, -0.5],
            [0.5, 0.0, -0.5],
            [0.5, 0.0, 0.5],
            [-0.5, 0.0, 0.5],
        ]
        indices = [
            [0, 1, 2],
            [0, 2, 3],
        ]
        super().__init__(vertices, indices, colors)


class OBJMesh(Mesh):
    def __init__(self, file: io.TextIOWrapper, colors: list[list[int]] = None) -> None:
        vertices = []
        indices = []

        with file:  # TODO: implement MTL parser, textures
            for line in file:
                parts = line.strip().split()
                if not parts:
                    continue
                if parts[0] == "v":
                    vertices.append([float(coord) for coord in parts[1:4]])
                elif parts[0] == "f":
                    buffer = [
                        int(part.split("/")[0])
                        -
                        1 for part in parts[1:]
                    ]
                    if len(buffer) == 3:
                        indices.append(buffer)
                    elif len(buffer) > 3:
                        for i in range(1, len(buffer) - 1):
                            indices.append(
                                [buffer[0], buffer[i], buffer[i + 1]])

        super().__init__(vertices, indices, colors)



import tkenginer as tke
import numpy as np
import time

from PIL import ImageDraw


class Demo(tke.Engine):
    def __init__(self) -> None:
        super().__init__("TkEnginer new demo")
        self.speed = 3
        self.sensetivity = 0.2
        self.last_mouse = None

    def update(self, delta: float) -> None:
        draw = ImageDraw.Draw(self.image)
        draw.text(
            (10, 10), 
            f"FPS: {int(1 / delta)}", 
            "white"
        )
        front, right, _ = tke.math.get_camera_vectors(self.yaw, self.pitch)

        mouse = self.get_mouse_position()
        if self.is_key_pressed("mouse_1"):
            self.yaw += (mouse[0] - self.last_mouse[0]) * self.sensetivity * delta
            self.pitch -= (mouse[1] - self.last_mouse[1]) * self.sensetivity * delta
            max_pitch = np.pi / 2 - 0.01
            self.pitch = max(-max_pitch, min(max_pitch, self.pitch))
        self.last_mouse = mouse

        if self.is_key_pressed("w"):
            self.position += front * self.speed * delta
        if self.is_key_pressed("a"):
            self.position -= right * self.speed * delta
        if self.is_key_pressed("s"):
            self.position -= front * self.speed * delta
        if self.is_key_pressed("d"):
            self.position += right * self.speed * delta


demo = Demo()
demo.scene = tke.Node(children=[
    tke.Node(
        mesh=tke.SphereMesh(8),
        transform=tke.Transform(
            position=[2.0, -3.0, -3.0],
            rotation=[np.pi/4, np.pi/2, np.pi/3],
            scale=[0.7, 0.5, 1.2]
        )
    ),
    tke.Node(
        mesh=tke.CubeMesh(
            colors=[
                [255, 0, 0, 255],
                [0, 255, 0, 255],
                [0, 0, 255, 255],
                [255, 255, 0, 255],
                [255, 0, 255, 255],
                [0, 255, 255, 255],
                [255, 165, 0, 255],
                [128, 0, 128, 255],
            ]
        ),
        transform=tke.Transform(
            position=[-5.0, 2.0, -3.0],
            rotation=[np.pi/6, np.pi/3, np.pi/4],
            scale=[1.5, 0.5, 1.0]
        )
    ),
    tke.RigidbodyNode(
        mesh=tke.PyramidWithSquareBaseMesh(),
        transform=tke.Transform(
            position=[4.0, 3.0, -3.0],
            rotation=[np.pi/3, np.pi/2, np.pi/5],
            scale=[0.8, 1.0, 0.5]
        ),
        mass=0.1
    ),
    tke.Node(
        mesh=tke.PyramidMesh(),
        transform=tke.Transform(
            position=[-2.0, -1.0, -3.0],
            rotation=[np.pi/4, np.pi/6, np.pi/2],
            scale=[1.0, 1.3, 0.6]
        )
    ),
    tke.Node(
        mesh=tke.ConeMesh(15),
        transform=tke.Transform(
            position=[1.0, 5.0, -3.0],
            rotation=[np.pi/2, np.pi/4, np.pi/6],
            scale=[1.0, 0.8, 1.5]
        )
    ),
    tke.Node(
        mesh=tke.PlaneMesh(),
        transform=tke.Transform(
            position=[0.0, -2.0, -3.0],
            rotation=[0.0, 0.0, 0.0],
            scale=[3.0, 3.0, 3.0]
        )
    ),
    # tke.GameObject(
    #     mesh=tke.OBJLoader(open("ak74.obj", "r")),
    #     transform=tke.Transform(
    #         position=[-1.0, 0.0, 3.0],
    #         rotation=[np.pi/4, np.pi/2, np.pi/3],
    #         scale=[0.1, 0.1, 0.1]
    #     )
    # )
])
demo.run()



import numpy as np

from .node import Node
from .transform import Transform
from .mesh import Mesh

gravity = np.array([0, -9.81, 0], dtype=np.float32)


class RigidbodyNode(Node):
    def __init__(
        self,
        mesh: Mesh = None,
        transform: Transform = None,
        children: list = None,
        mass: float = 1.0,
        use_gravity: bool = True,
    ) -> None:
        super().__init__(mesh, transform, children)
        self.velocity = np.zeros(3, dtype=np.float32)
        self.acceleration = np.zeros(3, dtype=np.float32)
        self.mass = mass
        self.use_gravity = use_gravity

    def apply_force(self, force: np.ndarray):
        self.acceleration += force / self.mass

    def update(self, delta: float) -> None:
        if self.use_gravity:
            self.acceleration += gravity
        self.velocity += self.acceleration * delta
        self.transform.position += self.velocity * delta
        self.acceleration[:] = 0



import copy

from .transform import *
from .mesh import *


class Node:
    def __init__(self, mesh: Mesh = None, transform: Transform = None, children: "Node" = None) -> None:
        self.mesh = mesh
        self.transform = transform if transform is not None else Transform()
        self.children = children if children is not None else list()

    def update(self, delta: float) -> None:
        pass

    def traverse(self, parent_transform=Transform()):
        global_transform = parent_transform @ self.transform
        yield self, global_transform
        for child in self.children:
            yield from child.traverse(global_transform)

